/*
 * Traditional system functions for filesystem access.
 *
 * Copyright (C) 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/types.h>

#include <errno.h>
#include <fcntl.h>

#include <fsclient/client.h>
#include "sys_client.h"



/* File descriptors.
   NOTE: These should reside in an extensible structure. */

#define desc_limit 64

static file_descriptor_t descs[desc_limit];

/* Allocation of descriptors, skipping the standard streams. */

static int start_offset = 3, next_free = 3, recover = 0;



/* Descriptor allocation and testing. */

static int find_free(void)
{
  int i;

  if (!recover && (next_free < desc_limit))
    return (next_free)++;

  if (next_free > desc_limit)
  {
    next_free = start_offset;
    recover = 1;
  }

  i = next_free;

  do
  {
    if (l4_is_invalid_cap(descs[i].server))
    {
      next_free = i + 1;
      return i + start_offset;
    }

    i++;

    if (i > desc_limit)
      i = start_offset;

  } while (i != next_free);

  return -1;
}

static int is_valid(int fd)
{
  if ((fd < 0) || (fd >= desc_limit))
    return 0;

  if (!recover && (fd >= next_free))
    return 0;

  if (l4_is_invalid_cap(descs[fd].server))
    return 0;

  return 1;
}



/* File operations. */

#define DEFAULT_BUFFER_SIZE 4096

int _close_r(struct _reent *reent, int fd)
{
  if (!is_valid(fd))
  {
    reent->_errno = EBADF;
    return -1;
  }

  client_flush(&descs[fd]);
  client_close(&descs[fd]);
  return 0;
}

int _fstat_r(struct _reent *reent, int fd, struct stat *buf)
{
  if (!is_valid(fd))
  {
    reent->_errno = EBADF;
    return -1;
  }

  return client_fstat(&descs[fd], buf);
}

int _isatty_r(struct _reent *reent, int fd)
{
  return 0;
}

int _link_r(struct _reent *reent, const char *oldpath, const char *newpath)
{
  return -1;
}

off_t _lseek_r(struct _reent *reent, int fd, off_t offset, int whence)
{
  if (!is_valid(fd))
  {
    reent->_errno = EBADF;
    return -1;
  }

  return client_seek(&descs[fd], offset, whence);
}

int _open_r(struct _reent *reent, const char *name, int flags, int mode)
{
  int fd = find_free();

  if (fd == -1)
  {
    reent->_errno = EBADF;
    return -1;
  }

  if (client_open(&descs[fd], name, flags, 0) == NULL)
  {
    reent->_errno = EACCES; /* NOTE: should obtain error code */
    return -1;
  }

  return fd;
}

ssize_t _read_r(struct _reent *reent, int fd, void *buffer, size_t nbytes)
{
  if (!is_valid(fd))
  {
    reent->_errno = EBADF;
    return -1;
  }

  return client_read(&descs[fd], buffer, nbytes);
}

int _rename_r(struct _reent *reent, const char *oldpath, const char *newpath)
{
  return -1;
}

int _stat_r(struct _reent *reent, const char *pathname, struct stat *buf)
{
  int fd = _open_r(reent, pathname, O_RDONLY, 0);
  int result;

  if (fd == -1)
    return -1;

  result = _fstat_r(reent, fd, buf);
  _close_r(reent, fd);

  return result;
}

int _unlink_r(struct _reent *reent, const char *pathname)
{
  return -1;
}

ssize_t _write_r(struct _reent *reent, int fd, const void *buffer, size_t nbytes)
{
  if (!is_valid(fd))
  {
    reent->_errno = EBADF;
    return -1;
  }

  return client_write(&descs[fd], buffer, nbytes);
}
